<?php

declare(strict_types=1);

namespace Drupal\Tests\project_browser\Kernel;

use Composer\InstalledVersions;
use Drupal\Component\FileSystem\FileSystem;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Recipe\Recipe;
use Drupal\KernelTests\KernelTestBase;
use Drupal\project_browser\EnabledSourceHandler;
use Drupal\project_browser\Plugin\ProjectBrowserSourceManager;
use Drupal\project_browser\ProjectType;
use Symfony\Component\Filesystem\Filesystem as SymfonyFilesystem;
use Symfony\Component\Finder\Finder;
use Symfony\Component\Finder\SplFileInfo;

/**
 * Tests the source plugin that exposes locally installed recipes.
 *
 * @group project_browser
 * @covers \Drupal\project_browser\Plugin\ProjectBrowserSource\Recipes
 */
class RecipesSourceTest extends KernelTestBase {

  /**
   * {@inheritdoc}
   */
  protected static $modules = ['project_browser', 'project_browser_test'];

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();

    if (!class_exists(Recipe::class)) {
      $this->markTestSkipped('This test cannot be run because the recipe system is not available.');
    }
    $this->installSchema('project_browser_test', [
      'project_browser_projects',
      'project_browser_categories',
    ]);
    $this->installConfig('project_browser_test');
    $this->installConfig('project_browser');
  }

  /**
   * @covers \project_browser_install
   * @covers \project_browser_project_browser_source_info_alter
   */
  public function testRecipeSourceIsEnabledAtInstallTime(): void {
    $this->assertNotContains('recipes', $this->config('project_browser.admin_settings')->get('enabled_sources'));

    $this->container->get(ModuleHandlerInterface::class)
      ->loadInclude('project_browser', 'install');
    project_browser_install();
    $this->assertContains('recipes', $this->config('project_browser.admin_settings')->get('enabled_sources'));

    $enabled_sources = $this->container->get(EnabledSourceHandler::class)
      ->getCurrentSources();
    $this->assertArrayHasKey('recipes', $enabled_sources);
  }

  /**
   * Tests that recipes are discovered by the plugin.
   */
  public function testRecipesAreDiscovered(): void {
    // Generate a fake recipe in the temporary directory.
    $generated_recipe_name = uniqid();
    $generated_recipe_dir = FileSystem::getOsTemporaryDirectory() . '/' . $generated_recipe_name;
    mkdir($generated_recipe_dir);
    file_put_contents($generated_recipe_dir . '/composer.json', '{"name": "drupal/bogus_recipe"}');
    file_put_contents($generated_recipe_dir . '/recipe.yml', 'name: Bogus');

    // Find the place where recipes are installed; this contributed recipe is
    // one of our dev dependencies.
    $installed_recipes_dir = InstalledVersions::getInstallPath('kanopi/imagemagick-configuration');
    $installed_recipes_dir = realpath($installed_recipes_dir);
    $installed_recipes_dir = dirname($installed_recipes_dir);
    $file_system = new SymfonyFilesystem();
    // Symlink the fake recipe into the place where the source plugin will
    // search, to prove that the plugin follows symlinks.
    $file_system->symlink($generated_recipe_dir, $installed_recipes_dir . '/' . $generated_recipe_name);

    $finder = Finder::create()
      ->in($this->getDrupalRoot() . '/core/recipes')
      ->directories()
      ->notName('example')
      ->depth(0);
    $expected_recipe_names = array_map(fn (SplFileInfo $dir) => $dir->getBasename(), iterator_to_array($finder));
    $expected_recipe_names[] = 'imagemagick-configuration';
    $expected_recipe_names[] = $generated_recipe_name;

    /** @var \Drupal\project_browser\ProjectBrowser\ProjectsResultsPage $projects */
    $projects = $this->container->get(ProjectBrowserSourceManager::class)
      ->createInstance('recipes')
      ->getProjects();
    $found_recipes = [];
    foreach ($projects->list as $project) {
      $this->assertNotEmpty($project->title);
      $this->assertSame(ProjectType::Recipe, $project->type);
      $found_recipes[$project->machineName] = $project;
    }
    $found_recipe_names = array_keys($found_recipes);

    // The `example` recipe (from core) should always be hidden.
    $this->assertNotContains('example', $expected_recipe_names);

    sort($expected_recipe_names);
    sort($found_recipe_names);
    $this->assertSame($expected_recipe_names, $found_recipe_names);

    // Ensure the package names are properly resolved.
    $this->assertSame('drupal/core', $found_recipes['standard']?->packageName);
    $this->assertSame('kanopi/imagemagick-configuration', $found_recipes['imagemagick-configuration']?->packageName);

    // The core recipes should have descriptions, which should become the body
    // text of the project.
    $this->assertArrayHasKey('standard', $found_recipes);
    // The need for reflection sucks, but there's no way to introspect the body
    // on the backend.
    $body = (new \ReflectionProperty($found_recipes['standard'], 'body'))
      ->getValue($found_recipes['standard']);
    $this->assertNotEmpty($body);

    // Clean up.
    $file_system->remove([
      $installed_recipes_dir . '/' . $generated_recipe_name,
      $generated_recipe_dir,
    ]);
  }

  /**
   * Tests that discovered recipes are limited by an allow-list.
   */
  public function testAllowList(): void {
    $expected_recipe_names = ['document_media_type', 'user_picture'];

    $this->config('project_browser.admin_settings')
      ->set('allowed_projects', [
        'recipes' => ['example', ...$expected_recipe_names],
      ])
      ->save();

    /** @var \Drupal\project_browser\ProjectBrowser\ProjectsResultsPage $projects */
    $projects = $this->container->get(ProjectBrowserSourceManager::class)
      ->createInstance('recipes')
      ->getProjects();
    $found_recipe_names = array_column($projects->list, 'machineName');

    // The `example` recipe (from core) should always be hidden, even if it's in
    // the allow list.
    $this->assertNotContains('example', $found_recipe_names);

    sort($expected_recipe_names);
    sort($found_recipe_names);
    $this->assertSame($expected_recipe_names, $found_recipe_names);
  }

}
